#include <stdlib.h>
#include <windows.h>
#include "resource.h"
const int AllFrame = 10;
const int ObjectCX = 80;
const int ObjectCY = 80;
const int MaxObject = 4;

struct OBJECTFRAME {
	BOOL bEnabled;
	int x;
	int y;
};

HBITMAP g_hBackBufBMP, g_hFrameBufBMP, g_hObjectBMP[AllFrame];
HDC g_hBackBufDC, g_hFrameBufDC, g_hObjectDC[AllFrame];
HRGN g_ObjectRgn[AllFrame];
SIZE ScreenSize;
HWND g_hWnd;
OBJECTFRAME ObjectFrame[AllFrame][MaxObject];

HRGN getRgnFromDC(HDC, HBITMAP);
int CreateBitmapBuffer(HINSTANCE, HDC *, HDC *, HBITMAP *, TCHAR [32]);
int CreateBuffer(HDC *, HDC *, HBITMAP *, int, int);
void ScreenClick(int, int);
void PaintObject(HDC, int, int, int);
void NextFrame();

HRGN getRgnFromDC(HDC hDC, HBITMAP hBitmap) {
	BITMAP bi;
	GetObject(hBitmap, sizeof(BITMAP), &bi);
	HRGN hRgn = CreateRectRgn(0, 0, 0, 0);

	HBITMAP hDIBmpBMP;
	HDC hDIBmpDC;
	BYTE* chPixelBit;
	chPixelBit = (BYTE*)malloc(bi.bmWidth * bi.bmHeight * 4);
	BITMAPINFO bmpInfo;
	bmpInfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmpInfo.bmiHeader.biWidth = bi.bmWidth;
	bmpInfo.bmiHeader.biHeight = bi.bmHeight;
	bmpInfo.bmiHeader.biPlanes = 1;
	bmpInfo.bmiHeader.biBitCount = 32;
	bmpInfo.bmiHeader.biCompression = BI_RGB;
	hDIBmpBMP = CreateDIBSection(NULL, &bmpInfo, DIB_RGB_COLORS, (VOID**)&chPixelBit, NULL, 0);
	hDIBmpDC = CreateCompatibleDC(GetDC(g_hWnd));
	SelectObject(hDIBmpDC, hDIBmpBMP);
	BitBlt(hDIBmpDC, 0, 0, ObjectCX, ObjectCY, hDC, 0, 0, SRCCOPY);

	int iCount = 0, iXPos, iYPos, iLine = 1;
	const int WIDTH_BYTE = 4 * ((bi.bmWidth * bmpInfo.bmiHeader.biBitCount + 31) / 32);
	for (iYPos = bi.bmHeight - 1; iYPos >= 0; iYPos--) {
		int FirstTruePx = -1;
		for (iXPos = 0; ; iXPos++) {
			if (iXPos > (bmpInfo.bmiHeader.biWidth - 1)) {
				iCount = iLine * WIDTH_BYTE;
				iLine++;
				break;
			}
			if (RGB(chPixelBit[iCount + 2], chPixelBit[iCount + 1], chPixelBit[iCount]) != RGB(0xFF, 0, 0xFF)) {
				if (FirstTruePx == -1) FirstTruePx = iXPos;
			}else {
				if (FirstTruePx != -1) {
					HRGN DiffRgn = CreateRectRgn(FirstTruePx, iYPos, iXPos, iYPos + 1);
					CombineRgn(hRgn, hRgn, DiffRgn, RGN_OR);
					DeleteObject(DiffRgn);
					FirstTruePx = -1;
				}
			}
			iCount += 4;
		}
	}
	DeleteObject(hDIBmpDC);
	DeleteObject(hDIBmpBMP);
	return hRgn;
}

int CreateBitmapBuffer(HINSTANCE hInstance, HDC *CompatibleDC, HDC *BufferDC, HBITMAP *BufferBMP, TCHAR BMPName[32]){
	HBITMAP TempBMP = LoadBitmap(hInstance ,BMPName);
	if(!TempBMP){
		return 1;
	}
	BITMAP bi;
	GetObject(TempBMP, sizeof(BITMAP), &bi );
	if(CreateBuffer(CompatibleDC, BufferDC, BufferBMP, bi.bmWidth, bi.bmHeight)) return 1;
	HDC TempDC = CreateCompatibleDC(*BufferDC);
	if(!TempDC){
		return 1;
	}
	if(!SelectObject(TempDC, TempBMP)){
		return 1;
	}

	BitBlt(*BufferDC, 0, 0, bi.bmWidth, bi.bmHeight, TempDC, 0, 0, SRCCOPY);
	
	DeleteObject(TempDC);
	DeleteObject(TempBMP);

	return 0;
}

int CreateBuffer(HDC *CompatibleDC, HDC *BufferDC, HBITMAP *BufferBMP, int cx, int cy){
	*BufferBMP = CreateCompatibleBitmap(*CompatibleDC, cx, cy);
	if(!*BufferBMP){
		return 1;
	}
	*BufferDC = CreateCompatibleDC(*CompatibleDC);
	if(!*BufferDC){
		return 1;
	}
	if(!SelectObject(*BufferDC, *BufferBMP)){
		return 1;
	}

	return 0;
}

void ScreenClick(int MouseX, int MouseY) {
	int x = MouseX - (ObjectCX / 2);
	int y = MouseY - (ObjectCY / 2);
	int i;
	for (i = 0; (i < MaxObject) && (ObjectFrame[0][i].bEnabled == 1); i++) {}
	if (i >= (MaxObject - 1)) return;
	ObjectFrame[1][i].bEnabled = 1;
	ObjectFrame[1][i].x = x;
	ObjectFrame[1][i].y = y;
}

void PaintObject(HDC hDC, int ObjIndex, int x, int y) {
	HRGN OffRgn;
	OffRgn = CreateRectRgn(0, 0, 0, 0);
	CombineRgn(OffRgn, g_ObjectRgn[ObjIndex], NULL, RGN_COPY);
	OffsetRgn(OffRgn, x, y);
	SelectClipRgn(hDC, OffRgn);
	BitBlt(hDC, x, y, ObjectCX, ObjectCY, g_hObjectDC[ObjIndex], 0, 0, SRCCOPY);
	SelectClipRgn(hDC, NULL);
	DeleteObject(OffRgn);
}

void NextFrame() {
	BOOL bDraw = 0;
	for (int i = (AllFrame - 1); i >= 0; i--) {
		for (int j = 0; j < MaxObject; j++) {
			if (ObjectFrame[i][j].bEnabled) {
				bDraw = 1;
				break;
			}
		}
		if (bDraw) break;
	}
	if (bDraw) {
		HRGN hFrameBufRgn;
		hFrameBufRgn = CreateRectRgn(0, 0, 0, 0);
		for (int i = (AllFrame - 1); i >= 0; i--) {
			for (int j = 0; j < MaxObject; j++) {
				if (ObjectFrame[i][j].bEnabled) {
					HRGN hObjectRectRgn;
					hObjectRectRgn = CreateRectRgn(ObjectFrame[i][j].x, ObjectFrame[i][j].y, ObjectCX + ObjectFrame[i][j].x, ObjectCY + ObjectFrame[i][j].y);
					CombineRgn(hFrameBufRgn, hFrameBufRgn, hObjectRectRgn, RGN_OR);
					DeleteObject(hObjectRectRgn);
				}
			}
		}
		SelectClipRgn(g_hFrameBufDC, hFrameBufRgn);
		BitBlt(g_hFrameBufDC, 0, 0, ScreenSize.cx, ScreenSize.cy, g_hBackBufDC, 0, 0, SRCCOPY);
		SelectClipRgn(g_hFrameBufDC, NULL);
		for (int i = (AllFrame - 1); i >= 0; i--) {
			for (int j = 0; j < MaxObject; j++) {
				if (ObjectFrame[i][j].bEnabled) {
					PaintObject(g_hFrameBufDC, i, ObjectFrame[i][j].x, ObjectFrame[i][j].y);
					if (i == AllFrame - 1) {
						PaintObject(g_hBackBufDC, i, ObjectFrame[i][j].x, ObjectFrame[i][j].y);
					}
				}
			}
			if (i != AllFrame - 1) memcpy(ObjectFrame[i + 1], ObjectFrame[i], sizeof(OBJECTFRAME) * MaxObject);
		}
		ZeroMemory(ObjectFrame[0], sizeof(OBJECTFRAME));
		HDC wDC = GetDC(g_hWnd);
		SelectClipRgn(wDC, hFrameBufRgn);
		BitBlt(wDC, 0, 0, ScreenSize.cx, ScreenSize.cy, g_hFrameBufDC, 0, 0, SRCCOPY);
		SelectClipRgn(wDC, NULL);
		DeleteObject(hFrameBufRgn);
	}
}

LRESULT CALLBACK WndProc(
	HWND hWnd, 
	UINT msg, 
	WPARAM wParam, 
	LPARAM lParam
){
	PAINTSTRUCT ps;
	HDC wDC;
	switch(msg){
	case WM_PAINT:
		wDC = BeginPaint(hWnd , &ps);
		BitBlt(wDC, 0, 0, ScreenSize.cx, ScreenSize.cy, g_hFrameBufDC, 0 , 0 , SRCCOPY);
		EndPaint(hWnd , &ps);
		break;
	case WM_TIMER:
		NextFrame();
		break;
	case WM_CREATE:
		ZeroMemory(ObjectFrame, sizeof(OBJECTFRAME) * AllFrame * MaxObject);
		g_hWnd = hWnd;
		wDC = GetDC(hWnd);
		CreateBuffer(&wDC, &g_hFrameBufDC, &g_hFrameBufBMP, ScreenSize.cx, ScreenSize.cy);
		CreateBuffer(&g_hFrameBufDC, &g_hBackBufDC, &g_hBackBufBMP, ScreenSize.cx, ScreenSize.cy);
		for(int i=0; i < AllFrame; i++){
			TCHAR strTemp[8];
			wsprintf(strTemp, TEXT("OBJ%02d"), 1 + i);
			CreateBitmapBuffer(((LPCREATESTRUCT)lParam)->hInstance, &g_hFrameBufDC, &g_hObjectDC[i], &g_hObjectBMP[i], strTemp);
			g_ObjectRgn[i] = getRgnFromDC(g_hObjectDC[i], g_hObjectBMP[i]);
		}
		BitBlt(g_hBackBufDC, 0, 0, ScreenSize.cx, ScreenSize.cy, CreateDC(TEXT("DISPLAY") , NULL , NULL , NULL), 0, 0, SRCCOPY);
		BitBlt(g_hFrameBufDC, 0, 0, ScreenSize.cx, ScreenSize.cy, g_hBackBufDC, 0, 0, SRCCOPY);
		SetTimer(hWnd , 1 , 100 , NULL);
		PostMessage(hWnd, WM_TIMER, 0, 0);
		break;
	case WM_KEYDOWN:
		if (wParam != VK_ESCAPE){
			break;
		}
	case WM_DESTROY:
		for (int i = 0; i < AllFrame; i++) {
			DeleteDC(g_hObjectDC[i]);
			DeleteObject(g_hObjectBMP[i]);
			DeleteObject(g_ObjectRgn[i]);
		}
		DeleteDC(g_hBackBufDC);
		DeleteObject(g_hBackBufBMP);
		DeleteDC(g_hFrameBufDC);
		DeleteObject(g_hFrameBufBMP);
		PostQuitMessage(0);
		break;
	case WM_LBUTTONDOWN:
		ScreenClick(LOWORD(lParam), HIWORD(lParam));
		break;
	default:
		return DefWindowProc(hWnd, msg, wParam, lParam);
	}
	return 0;
}
int WinMain(
	HINSTANCE hInstance, 
	HINSTANCE hPrevInstance, 
	LPTSTR lpCmdLine, 
	int nCmdShow
){
	if(FindWindow(TEXT("BOMB_CLASS"), TEXT("BOMB"))) return 0;
	SetThreadPriority(GetCurrentThread(), 4);
	ScreenSize.cx = GetSystemMetrics(SM_CXSCREEN);
	ScreenSize.cy = GetSystemMetrics(SM_CYSCREEN);

	WNDCLASS wc;
	ZeroMemory(&wc, sizeof(WNDCLASS));
	wc.style = CS_HREDRAW | CS_VREDRAW;
	wc.lpfnWndProc = WndProc;
	wc.hInstance = hInstance;
	wc.lpszClassName = TEXT("BOMB_CLASS");
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	if(!RegisterClass(&wc)){
		return 1;
	}
	HWND hWnd = CreateWindowEx(
		WS_EX_TOPMOST | WS_EX_TOOLWINDOW,
		TEXT("BOMB_CLASS"), 
		TEXT("BOMB"),
		WS_POPUP ,
		0, 0 , ScreenSize.cx, ScreenSize.cy,
		0,
		0,
		hInstance,
		0
	);
	if(!hWnd){
		return 1;
	}
	SetThreadPriority(GetCurrentThread(), 3);
	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);
	MSG msg;
	while(GetMessage(&msg, 0, 0, 0) > 0){
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return (int)msg.wParam;
}
